#pragma once

#include <cassert>
#include <optional>

template<typename T>
class MinuVektor {
public:
    MinuVektor() {
        m_Buffer = new T[m_Capacity];
    }

    explicit MinuVektor(size_t capacity)
        : m_Capacity(capacity) {
        m_Buffer = new T[capacity];
    }

    MinuVektor(size_t capacity, T initial)
        : m_Size(capacity), m_Capacity(capacity) {
        m_Buffer = new T[capacity];
        for (size_t i = 0; i < m_Capacity; i++) {
            m_Buffer[i] = initial;
        }
    }

    ~MinuVektor() {
        delete[] m_Buffer;
    }

    void push_back(T value) {
        if (m_Size >= m_Capacity) {
            resize(m_Capacity * 2);
        }

        m_Buffer[m_Size++] = value;
    }

    std::optional<T> get(size_t idx) {
        if (idx >= m_Size) {
            return std::nullopt;
        }

        return m_Buffer[idx];
    }

    std::optional<T> front() {
        return get(0);
    }

    std::optional<T> back() {
        return get(m_Size - 1);
    }

    T operator[](size_t idx) {
        assert(idx < m_Size);
        return *get(idx);
    }

    template<typename U = T>
    requires std::integral<U>
    static MinuVektor<U> iota(size_t sz) {
        MinuVektor v(sz);
        for (size_t i = 0; i < sz; i++) {
            v.push_back(i);
        }
        return v;
    }

    template<typename U>
    MinuVektor<U> transform(auto fn) const {
        MinuVektor<U> v(size());
        for (auto i = 0u; i < size(); i++) {
            v.push_back(fn(m_Buffer[i]));
        }
        return v;
    }

    [[nodiscard]] const T *buf() const { return m_Buffer; }
    [[nodiscard]] auto size() const { return m_Size; }
    [[nodiscard]] auto capacity() const { return m_Capacity; }

private:
    void resize(const size_t newCapacity) {
        if (newCapacity < m_Capacity) {
            return;
        }

        T *newBuffer = new T[newCapacity];
        for (size_t i = 0; i < m_Capacity; i++) {
            newBuffer[i] = m_Buffer[i];
        }
        delete[] m_Buffer;
        m_Buffer = newBuffer;
        m_Capacity = newCapacity;
    }

private:
    size_t m_Size{0};
    size_t m_Capacity{2};
    T *m_Buffer = nullptr;
};

template<typename T>
std::ostream &operator<<(std::ostream &os, MinuVektor<T> &v) {
    os << '{';
    for (size_t i = 0; i < v.size(); i++) {
        os << v[i];
        if (i != v.size() - 1) {
            os << ", ";
        }
    }
    os << '}';
    return os;
}